print("import: 'gevent'")
import gevent

print("import: 'gevent.libev'")
import gevent.libev

print("import: 'gevent.libuv'")
import gevent.libuv

print("import: 'gevent.resolver'")
import gevent.resolver

